// Gary Yngve
// 10/1/2005
// This class provides all the state and behavior needed for Minesweeper.
// Note how public methods are extremely limited in capability, and the
// only public method that breaks the rules of the game has "cheating"
// associated with it.

import java.io.*;
import java.util.*;

public class MineBoard {

    // any (i,j) params in this class refer to Cartesian coordinates,
    // i.e. lower left is (0,0), i goes right, j goes up

    private int w, h; // width and height of board
    private MineCell[] board; // internal format only affects get/set
    private int numMines; // number of mines on the board
    private int numExplored; // number of cells safely explored on the board
    private int numMarked; // number of cells marked, whether right or wrong
    private boolean gameLost; // step on a mine?
    private boolean cheating; // did they access any cheat methods?

    // static factory methods for the standard Minesweeper levels
    public static MineBoard smallFactory() {
        return new MineBoard(8,8,10);
    }

    public static MineBoard mediumFactory() {
        return new MineBoard(16,16,40);
    }

    public static MineBoard largeFactory() {
        return new MineBoard(30,16,99);
    }

    // makes a new w x h board with numMines mines, game initialized
    // (most vars default to 0/false and are not explicitly included)
    public MineBoard(int w, int h, int numMines) {
        if (numMines < 0 || numMines > w*h)
            throw new IllegalArgumentException("numMines " + numMines
                                          + "outside [0," + (w*h) + "]");
        this.w = w;
        this.h = h;
        this.numMines = numMines;
        board = new MineCell[w*h];
        reset();
    }

    public MineBoard(boolean[][] mines) {
        w = mines.length;
        h = mines[0].length;
        for(int i = 1; i < w; i++)
            if (mines[i].length != h)
                throw new IllegalArgumentException("2-d array not rectangular");
        board = new MineCell[w*h];
        for(int i = 0; i < w; i++)
            for(int j = 0; j < h; j++) {
                if (mines[i][j]) {
                    set(i, j,MineCell.UNEXPLOREDMINE);
                    numMines++;
                } else set(i,j,MineCell.UNEXPLOREDEMPTY);
            }
    }

    public MineBoard(String name) {
        Scanner sc;
        try {
            sc = new Scanner(new File(name));
        } catch (Exception e) { // checked exception...
            throw new RuntimeException("File not found: " + name);
        }
        if (sc.hasNextLine()) {
            Scanner s = new Scanner(sc.nextLine());
            if (s.hasNextInt()) {
                w = s.nextInt();
            } else throw new RuntimeException("bad format");
            if (s.hasNextInt()) {
                h = s.nextInt();
            } else throw new RuntimeException("bad format");
        }
        board = new MineCell[w*h];
        for(int j = 0; j < h; j++) {
            if (!sc.hasNextLine()) throw new RuntimeException("bad format");
            String s = sc.nextLine();
            if (s.length() < w) throw new RuntimeException("bad format");
            for(int i = 0; i < w; i++) {
                if (s.charAt(i) == '*') {
                    set(i,j,MineCell.UNEXPLOREDMINE);
                    numMines++;
                }
                else set(i,j,MineCell.UNEXPLOREDEMPTY);
            }
        }
    }

    public void reset() {
        populateBoard();
        numExplored = numMarked = 0;
        gameLost = cheating = false;
        printMines();
    }

    // populates the w x h grid with mines, uniformly randomly.
    // rejection method:  poorer performance if numMines > w*h/2
    private void populateBoard() {
        Random r = new Random(); // ok, since not spawning many at once
        for(int i = 0; i < w; i++)
            for(int j = 0; j < h; j++)
                set(i,j,MineCell.UNEXPLOREDEMPTY); // clear the board
        for(int k = 0; k < numMines; k++) {
            int i, j;
            do { // pick a spot
                i = r.nextInt(w);
                j = r.nextInt(h);
            } while (isMine(i,j)); // keep trying if already occupied
            set(i,j,MineCell.UNEXPLOREDMINE); // unoccupied, so place it
        }
    }

    private boolean isMine(int i, int j) {
        return get(i,j) == MineCell.MARKEDMINE
            || get(i,j) == MineCell.UNEXPLOREDMINE;
    }

    // the grossness of the internal representation of board is confined
    // to just get and set, which are private.  (i,j) -> j*w+i, i<w, j<h, i,j>=0
    // Returns OOB if not (i,j) not in grid.
    private MineCell get(int i, int j) {
        if (!inBounds(i,j))
            return MineCell.OOB;
        return board[j*w+i];
    }

    // set is the internal, powerful way to modify a cell on the board.
    // all public methods that modify the board (in a restricted fashion)
    // will end up calling set at some point.
    // Throws IOOB exception if (i,j) not in grid.
    private void set(int i, int j, MineCell val) {
        if (!inBounds(i,j))
            throw new IndexOutOfBoundsException("OOB: (" + i + "," + j + ")"
                                                + "not in (0,0)-(" + w + "," + h + ")");
        board[j*w+i] = val;
    }

    // I'm following the paradigm of printing all needed debug info with
    // toString.  Exceptions are just printing pertinent info, but maybe
    // they should use toString too?
    public String toString() {
        String s = "board " + w + " x " + h + ", " + numMines + "mines, "
            + numMarked + " cells marked" + "\n";
        s += "num explored cells: " + numExplored + " out of " + numSafe()
            + ", gameLost=" + gameLost + ", cheating=" + cheating + "\n";
        s += MineCell.legendToString() + "\n";
        for(int j = 0; j < h; j++) {
            for(int i = 0; i < w; i++)
                s += get(i,j).toShortString();
            s += "\n";
        }
        return s;
    }

public void printMines() {
    System.out.println(w + " " + h);
    for(int j = 0; j < h; j++) {
        for(int i = 0; i < w; i++)
            if (isMine(i,j)) System.out.print("*");
            else System.out.print("-");
        System.out.println();
    }
}

    // sees if (i,j) is in the grid
    public boolean inBounds(int i, int j) {
        return i >= 0 && i < w && j >= 0 && j < h;
    }

    public int width() {
        return w;
    }

    public int height() {
        return h;
    }

    // number of safe cells to be cleared is total-numMines.
    // the game is won when numExplored == numSafe()
    public int numSafe() {
        return w*h-numMines;
    }

    public int numMarked() {
        return numMarked;
    }

    public int numMines() {
        return numMines;
    }

    public int numExplored() {
        return numExplored;
    }

    // game is over if hit a mine or if all safe squares are explored
    public boolean isGameOver() {
        return gameLost || numExplored - numSafe() == 0;
    }

    // need to check isGameOver first
    public boolean isWinner() {
        if (!isGameOver())
            throw new IllegalStateException("game not over yet");
        return !gameLost;
    }

    public boolean isCheater() {
        return cheating;
    }

    // only invoke this method if the game is over or you want to cheat.
    public boolean isMine_cheat(int i, int j) {
        if (!isGameOver()) cheating = true;
        return isMine(i,j);
    }

    // note that isMarked is different from unmarkable -- the latter also
    // checks if game over

    // Note that !isMarked does not mean markable, and !isExplored does not
    // mean explorable.  This is probably a bug waiting to happen, but I'm
    // trying to handle OOB as state rather than an exception, to make
    // neighborhood queries near borders easier.

    // unknown if it's not OOB or CLEAREDEMPTY
    public boolean isUnknown(int i, int j) {
        return get(i,j) == MineCell.UNEXPLOREDMINE ||
               get(i,j) == MineCell.UNEXPLOREDEMPTY ||
               get(i,j) == MineCell.MARKEDMINE ||
               get(i,j) == MineCell.MARKEDEMPTY;
    }

    public boolean isMarked(int i, int j) {
        return  get(i,j) == MineCell.MARKEDMINE ||
                get(i,j) == MineCell.MARKEDEMPTY;
    }

    public boolean isOOB(int i, int j) {
        return get(i,j) == MineCell.OOB;
    }

FILL IN YOUR CODE FROM HW0 HERE (8 METHODS)

}


