// Gary Yngve
// 10/1/2005
// This class provides both graphics and a user interface to
// go on top of the underlying MineBoard engine.
// It has a main to start everything up.

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

public class MineUI implements MouseListener {

// throughout this class, bp will refer to a point in board coordinates
// (0,0) -- (mb.width(),mb.height()), and wp will refer to a point in
// world coordinates, i.e. the window

    private MineBoard mb;
    private DrawingPanel dp;
    private Graphics2D gr;
    private Point lostbp; // where you lost (null until then)
    private static int MARGIN = 20; // height of header
    private static int CELLWIDTH = 20; // width of cell on screen
    private static int INSET = 4; // how many pixels to center text
    // right now, CELLWIDTH not incorporated with font sizes
    Color[] COUNTCOLORS = {new Color(0,0,255),   new Color(0,128,0),
                         new Color(128,128,0), new Color(128,0,128),
                         new Color(255,0,0),   new Color(128,0,0),
                         new Color(0,128,128), new Color(0,0,128)};
    // array indices 0..7 correspond to colors for number of mines 1..8

        // given the current board, sets up the gfx and mouse stuff
    public MineUI(MineBoard mb) {
        this.mb = mb;
        dp = new DrawingPanel(mb.width()*CELLWIDTH,
                                              mb.height()*CELLWIDTH+MARGIN);
        dp.addMouseListener(this);
        gr = (Graphics2D) dp.getGraphics();
        drawAll();
    }

// this method gets called on every mouse click
    public void mouseClicked(MouseEvent e) {
        int but = e.getButton();
        Point bp=worldToBoard(e.getPoint());
        if (mb.inBounds(bp.x,bp.y)) {
            if (but == e.BUTTON1) // left button
                explore(bp);
            else if (but == e.BUTTON2 || but == e.BUTTON3)
                toggleMark(bp); // middle or right button
        } else { // if OOB, reset the game (hack to reset by clicking header)
            mb.reset();
            drawAll();
        }
    }

// crap I need because I'm not extending MouseAdapter
    public void mouseEntered(MouseEvent e) {} //empty
    public void mouseExited(MouseEvent e) {} //empty
    public void mousePressed(MouseEvent e) {} //empty
    public void mouseReleased(MouseEvent e) {} //empty

    protected MineBoard mb() { 
	return mb;
    }


    protected DrawingPanel dp() { 
	return dp;
    }

// very useful functions convert wp->bp and bp->wp
    protected Point worldToBoard(Point wp) {
        return new Point(wp.x/CELLWIDTH,
                            (int)Math.floor(1.*(wp.y-MARGIN)/CELLWIDTH));
    }

    protected Point boardToWorld(Point bp) {
        return new Point(bp.x*CELLWIDTH,bp.y*CELLWIDTH+MARGIN);
    }

    protected void explore(Point bp) {
        if (mb.explorable(bp.x,bp.y)) { // check before doing
            mb.exploreCell(bp.x,bp.y);
            if (mb.isGameOver()) { // lost, reveal everything
                lostbp = bp;
                drawAll();
            }
            else { // if not game over, only need to update current square
                draw(bp);  // and header
                drawHeader();
            }
        }
    }

    protected void drawHeader() {
        gr.setFont(new Font(null,Font.BOLD,MARGIN-INSET));
        gr.setColor(Color.BLACK);
        gr.fillRect(0,0,CELLWIDTH*mb.width(),MARGIN);
        gr.setColor(Color.WHITE);
        String face;
        if (mb.isGameOver()) {
            if (mb.isWinner()) face = "8)";
            else face = ":(";
        } else face = ":)";
        gr.drawString(" " + (mb.numMines()-mb.numMarked()) + "     "
                 + (mb.numSafe()-mb.numExplored()) + "   " + face + "     reset",
                 INSET,MARGIN-INSET); // no guarantee of fitting in window
    }

    protected void toggleMark(Point bp) {
        if (mb.markable(bp.x,bp.y)) { // check before doing
            mb.markCell(bp.x,bp.y);
            draw(bp);
        }
        else if (mb.unmarkable(bp.x,bp.y)) {
            mb.unmarkCell(bp.x,bp.y);
            draw(bp);
        }
        drawHeader();
    }

    protected void drawAll() {
        for(int i = 0; i < mb.width(); i++)
            for(int j = 0; j < mb.height(); j++)
                draw(new Point(i,j));
         drawHeader();
    }

    protected void draw(Point bp) {
        Point wp=boardToWorld(bp);
        gr.setFont(new Font(null,Font.BOLD,CELLWIDTH-INSET));
        if (mb.isUnknown(bp.x,bp.y)) {
            if (mb.isGameOver() && bp.equals(lostbp))
                gr.setColor(Color.RED); // you hit the mine here
            else
                gr.setColor(new Color(200,192,184));
            gr.fill3DRect(wp.x,wp.y,CELLWIDTH,CELLWIDTH,true); // unexplored
            gr.setColor(Color.BLACK);
            if (mb.isMarked(bp.x,bp.y)) {
                gr.drawString("m",wp.x+INSET,wp.y+CELLWIDTH-INSET); // marked as mine
                if (mb.isGameOver() && !mb.isMine_cheat(bp.x,bp.y)) {
                    gr.setColor(Color.RED); // marked wrongly
                    gr.drawString("x",wp.x+INSET,wp.y+CELLWIDTH-INSET);
                }
            } else if (mb.isGameOver() && mb.isMine_cheat(bp.x,bp.y)) { // unmarked mine
                gr.drawString("*",wp.x+INSET,wp.y+CELLWIDTH-INSET);
            }
        } else if (mb.askable(bp.x,bp.y)) {
            gr.setColor(new Color(255,248,240));
            gr.fill3DRect(wp.x,wp.y,CELLWIDTH,CELLWIDTH,false);
            int i = mb.ask(bp.x,bp.y);
            if (i > 0) {
                gr.setColor(COUNTCOLORS[i-1]); // number of neighboring mines
                gr.drawString(Integer.toString(i),wp.x+INSET,wp.y+CELLWIDTH-INSET);
            }
        }
    }

    public static void main(String[] args) {
        javax.swing.SwingUtilities.invokeLater(new Runnable() {
            public void run() { // change here to get different size game
                new MineUI(MineBoard.smallFactory());
            }
        });
    }
}
