// Gary Yngve 10/10/2005
// SmartMineUI inherits most of its behavior from MineUI.
// But it automates exploring and marking, as long as it can logically
// deduce correct behavior.  Adjust SLEEPTIME to control the speed
// of the animation (when not hindered by the compute times of the strategies)
// Strategies include simple zero- and first-order strategies, as well as
// higher powered strategies obtained from constrained sampling.
// Finally an endgame analysis is invoked when few squares remain.

// ALL DEDUCTIONS ASSUME THAT A MARKED SQUARE IS A MINE.

import java.awt.Point;
import java.util.*;

public class SmartMineUI extends MineUI {
    private static int SLEEPTIME = 100; // (in milliseconds)
    private static int NUMSAMPLES = 100; // number of samples for DeriveRule
    private static int NUMENDSAMPLES = 100; // number of samples for endgame
    private static int ENDSIZE = 30; // when fewer unknown squares than
                                                           // ENDSIZE, solve the endgame

    private MineStrategy[] strategies; // array of MineStrategies to be tried
                                                             // in order (fastest to slowest)

    SmartMineUI(MineBoard mb) {
        super(mb); // invoke parent constructor
        strategies = new MineStrategy[] {
                               ZerosMineStrategy.INSTANCE,
                               CountMineStrategy.INSTANCE,
                               new DeriveRuleMineStrategy(NUMSAMPLES) };
    }

// a little bit of a hack -- drawHeader gets called in the parent every time
// that an explore or mark is done
    protected void drawHeader() {
        super.drawHeader();
        try { // nappy time
            Thread.sleep(SLEEPTIME);
        } catch (Exception e) { // some silly checked exception
            System.out.println("Dunno what this is for.");
        }
        dp().paintImmediately(); // force a paint, because we're not done
                // yet processing the MouseListener callback
    }

// we only need to decide strategies on a point that is useful
    protected boolean useful(Point bp) {
        return  (mb().explorable(bp.x,bp.y) && mb().markable(bp.x,bp.y));
    }

// sees if unknown squares < ENDSIZE and game still in progress
    protected boolean isEndGame() {
        return !mb().isGameOver() &&
           mb().width()*mb().height()-mb().numMarked()-mb().numExplored()
                < ENDSIZE;
    }

// solves the EndGame.  Because new information may be provided per endgame
// analysis, the analysis continues until no new discoveries are reported
// or game is over
    protected void solveEndGame() {
        boolean changes;
        do {
            List toExplore = new LinkedList(), toMark = new LinkedList();
            EndGame e = new EndGame(mb(),NUMENDSAMPLES);
            e.analyze(toExplore,toMark);
            changes = !toExplore.isEmpty() || !toMark.isEmpty();
            for(Iterator i = toExplore.iterator(); i.hasNext(); /*empty*/)
                super.explore((Point)i.next()); // invokes the parent's explore
                // mandatory cast because i.next returns Object
            for(Iterator i = toMark.iterator(); i.hasNext(); )
                super.toggleMark((Point)i.next()); // parent's toggleMark
        } while (changes && !mb().isGameOver()); // stop when gameOver too
    }

// adds the askable neighbors to the active list to be strategized.
// exposing a square can trigger a new decision in a square a distance
// up to two away, though we filter out some cases in which no decision
// can be made (a square surrounded by 8 unknowns)
    protected void addNeighbors(List activeList, Point bp) {
        for(int i = -2; i <= 2; i++)
            for(int j = -2; j <= 2; j++) {

                if (i<-1 || j<-1 || i>1 || j>1) {
                    boolean ask = false;
                    for(int k = -1; k <= 1; k++)
                        for(int l = -1; l <= 1; l++) {
                            if (mb().askable(bp.x+i+k,bp.y+j+l))
                                ask = true;
                        }
                    if (!ask) continue; // surrounded by 8 unknowns: skip
            }

            if (mb().isUnknown(bp.x+i,bp.y+j) && // isUnknown is a poor name
                !mb().isMarked(bp.x+i,bp.y+j)) // isUnexplored would be better
                activeList.add(new Point(bp.x+i,bp.y+j));
        }
    }

// the operation should be equivalent to a pop() or dequeue(),
// assuming that addNeigbors was equivalent to a push() or enqueue()
    protected Point removeNext(LinkedList activeList) {
        throw new RuntimeException("not implemented yet");
    }

// First explore explores the square (i,j) requested from the UI
// (explore invokes the parent's explore).
// Then explore establishes a list of active squares by adding
// the neighbors of (i,j).  (assuming stack): Squares are popped
// from the list.  If a decision can be made (mark/explore), it is executed,
// and the neighbors of the executed square are added to the list.
// Squares where a decision could not be reached are put on the nextList,
// to become the activeList for the next strategy, when all squares are
// tried for the current strategy.  If at a latter strategy and a square is
// explored/marked, revert back to first strategy.
// Because a location in the active list may not be useful to strategize about
// (game over or already marked/explored), it is checked prior to strategizing.
// Finally, solves the endgame if appropriate.
    protected void explore(Point bp) {
        throw new RuntimeException("not implemented yet");
    }

    public static void main(String[] args) {
        javax.swing.SwingUtilities.invokeLater(new Runnable() {
            public void run() { // change here to get different size game
    //        new SmartMineUI(MineBoard.largeFactory());
                new SmartMineUI(new MineBoard("mine3.txt"));
            }
        });
    }

}
